/*
	
	ViewStubs.cpp
	
	Copyright (c)1998 by Angry Red Planet.

	This code is distributed under a modified form of the
	Artistic License.  A copy of this license should have
	been included with it; if this wasn't the case, the
	entire package can be obtained at
	<URL:http://www.angryredplanet.com/>.

	Interfaces to various Be UI objects that need a bit
	of extra help...
*/

#ifndef ARPLAYOUT_VIEWSTUBS_H
#include "ArpLayout/ViewStubs.h"
#endif

#ifndef ARPKERNEL_ARPCOLOR_H
#include "ArpKernel/ArpColor.h"
#endif

#ifndef ARPKERNEL_ARPDEBUG_H
#include "ArpKernel/ArpDebug.h"
#endif

#ifndef _MENUITEM_H
#include <be/interface/MenuItem.h>
#endif

#ifndef _AUTOLOCK_H
#include <be/support/Autolock.h>
#endif

#include <float.h>
#include <algobase.h>
#include <string.h>

ArpMOD();

static BRect InitFrame(-10000,-10000,-9800,-9800);
//static BRect InitFrame(0,0,1,1);

/* ------ ArpBox object ------
 */
 
ArpBox::ArpBox(const char* name, const char* label,
				border_style border)
#ifdef IS_BOX
	: BBox(InitFrame, name, B_FOLLOW_NONE,
			B_WILL_DRAW //|B_FULL_UPDATE_ON_RESIZE
			|B_FRAME_EVENTS|B_NAVIGABLE_JUMP
			, //|B_FULL_UPDATE_ON_RESIZE,
			border),
	 ArpViewWrapper(this)
#else
	 : ArpLayoutable()
#endif
{
	PG_BasicFont = "BoldFont";
#ifdef IS_BOX
	OwnView();
	SetLabel(label);
#endif
	indent_l = indent_t = indent_r = indent_b = 0;
}

ArpBox::ArpBox(BMessage* data)
#ifdef IS_BOX
	: BBox(data), ArpViewWrapper(data, this)
#else
	: ArpLayoutable(data)
#endif
{
#ifdef IS_BOX
	OwnView();
#endif
	indent_l = indent_t = indent_r = indent_b = 0;
}

ArpBox* ArpBox::Instantiate(BMessage* archive)
{
	if ( validate_instantiation(archive, "ArpBox") ) 
		return new ArpBox(archive); 
	return NULL;
}

status_t ArpBox::Archive(BMessage* data, bool deep)
{
#ifdef IS_BOX
	status_t status = BBox::Archive(data,false);
	if( status == B_NO_ERROR )
		status = ArpLayoutable::Archive(data,deep);
#else
	status_t status = ArpLayoutable::Archive(data,deep);
#endif
	if( status == B_NO_ERROR )
		status = data->AddString("class","ArpBox");
	return status;
}

#ifdef IS_BOX
void ArpBox::Draw(BRect updateRect)
{
	ArpD(cdb << "Drawing Box " << Name() << " at " << updateRect << endl);
	inherited::Draw(updateRect);
	DrawLayout(this, updateRect);
}

#endif

void ArpBox::ComputeDimens(void)
{
	//SetLabel(Label());
#ifdef IS_BOX
	ArpViewWrapper::ComputeDimens();
#else
	inherited::ComputeDimens();
#endif
	
	float fw = PV_BasicFont.StringWidth("W");
	font_height fhs;
	PV_BasicFont.GetHeight(&fhs);
	float fh = fhs.ascent+fhs.descent+fhs.leading;
#ifdef IS_BOX
	float ps = PenSize();
#else
	float ps = 1;
#endif
	if( ps < 1 ) ps = 1;
	indent_l = indent_r = floor( (fw/2) + (ps*2) + .5 );
	indent_b = floor( (fh/2) + (ps*2) + .5 );

#ifdef IS_BOX
	const char* lab = Label();
#else
	const char* lab = NULL;
#endif
	float min_w = 0;
	if( lab ) {
		min_w = PV_BasicFont.StringWidth(lab)
				+ indent_l + indent_r;
		indent_t = floor( fh + ps*2 + .5 );
	} else {
		indent_t = floor( (fh/2) + ps*2 + .5 );
	}
	cur_dimens.min_width += indent_l + indent_r;
	cur_dimens.min_height += indent_t + indent_b;
	cur_dimens.pref_width += indent_l + indent_r;
	cur_dimens.pref_height += indent_t + indent_b;
	cur_dimens.max_width += indent_l + indent_r;
	cur_dimens.max_height += indent_t + indent_b;
	if( cur_dimens.max_width > ArpAnySize ) cur_dimens.max_width = ArpAnySize;
	if( cur_dimens.max_height > ArpAnySize ) cur_dimens.max_height = ArpAnySize;
	//cur_dimens.max_width = cur_dimens.max_height = FLT_MAX/2;
	cur_dimens.min_width = max(cur_dimens.min_width,min_w);
	cur_dimens.pref_width = max(cur_dimens.pref_width,min_w);
}

void ArpBox::Layout(void)
{
	BRect frm = LayoutBounds();
	int32 num = CountLayoutChildren();

	frm.left += indent_l;
	frm.top += indent_t;
	frm.right -= indent_r;
	frm.bottom -= indent_b;
	
	if( num >= 1 ) {
		ArpLayoutable* child = LayoutChildAt(0);
		ArpD(cdb << ADH << "Moving child to " << frm << endl);
		if( child ) child->SetLayout(frm);
	}	
}

/* ------ ArpButton object ------
 */
 
ArpButton::ArpButton(const char* name, const char* label,
						BMessage* message)
	:BButton(InitFrame, name, label, message, B_FOLLOW_NONE,
				B_WILL_DRAW|B_NAVIGABLE|B_FULL_UPDATE_ON_RESIZE),
	 ArpViewWrapper(this)
{
	OwnView();
	SetLayoutName(name);
}

ArpButton::ArpButton(BMessage* data)
	: BButton(data), ArpViewWrapper(data, this)
{
	OwnView();
}

ArpButton* ArpButton::Instantiate(BMessage* archive)
{
	if ( validate_instantiation(archive, "ArpButton") ) 
		return new ArpButton(archive); 
	return NULL;
}

status_t ArpButton::Archive(BMessage* data, bool deep)
{
	status_t status = BButton::Archive(data,false);
	if( status == B_NO_ERROR )
		status = ArpLayoutable::Archive(data,deep);
	if( status == B_NO_ERROR )
		status = data->AddString("class","ArpButton");
	return status;
}

void ArpButton::ComputeDimens(void)
{
	SetLabel(Label());
	ArpViewWrapper::ComputeDimens();
}

/* ------ ArpListView object ------
 */
 
ArpListView::ArpListView(const char* name, list_view_type type)
	:BListView(InitFrame, name, type, B_FOLLOW_NONE),
	 ArpViewWrapper(this)
{
	PG_BackgroundColor = "FillBackColor";
	PG_ForegroundColor = "FillTextColor";
	OwnView();
}

ArpListView::ArpListView(BMessage* data)
	: BListView(data), ArpViewWrapper(data, this)
{
	OwnView();
}

ArpListView* ArpListView::Instantiate(BMessage* archive)
{
	if ( validate_instantiation(archive, "ArpListView") ) 
		return new ArpListView(archive); 
	return NULL;
}

status_t ArpListView::Archive(BMessage* data, bool deep)
{
	status_t status = BListView::Archive(data,false);
	if( status == B_NO_ERROR )
		status = ArpLayoutable::Archive(data,deep);
	if( status == B_NO_ERROR )
		status = data->AddString("class","ArpListView");
	return status;
}

void ArpListView::GetPreferredSize(float* w, float* h)
{
	if( w ) *w = 0;
	if( h ) *h = 0;
}

void ArpListView::MakeFocus(bool focusState)
{
	BListView::MakeFocus(focusState);
	ArpLayoutable::SetFocusShown(focusState);
}

void ArpListView::ComputeDimens(void)
{
	ArpViewWrapper::ComputeDimens();
	float min_w=0, min_h=0;
	float pref_w=0, pref_h=0;
	
	int32 num = CountItems();
	int32 max_num = 4;
	for( int32 i=0; i<num; i++ ) {
		BListItem* item = ItemAt(i);
		if( item ) {
			item->Update(this,&PV_BasicFont);
			float w = item->Width();
			float h = item->Height();
			min_h = max(min_h,h);
			pref_w = max(pref_w,w);
			if( max_num > 0 ) pref_h += h;
			max_num--;
		}
	}
	pref_h+=2;
	float fw = PV_BasicFont.StringWidth("WWWW");
	font_height fhs;
	PV_BasicFont.GetHeight(&fhs);
	float fh = fhs.ascent+fhs.descent+fhs.leading;
	min_w = max(min_w,fw);
	min_h = max(min_h,fh);
	pref_w = max(pref_w,min_w);
	pref_h = max(pref_h,min_h);
	cur_dimens.min_width = max(cur_dimens.min_width,min_w);
	cur_dimens.min_height = max(cur_dimens.min_height,min_h);
	cur_dimens.pref_width = max(cur_dimens.pref_width,pref_w);
	cur_dimens.pref_height = max(cur_dimens.pref_height,pref_h);
	cur_dimens.max_width = cur_dimens.max_height = ArpAnySize;
}

/* ------ ArpOutlineListView object ------
 */
 
ArpOutlineListView::ArpOutlineListView(const char* name, list_view_type type)
	:BOutlineListView(InitFrame, name, type, B_FOLLOW_NONE),
	 ArpViewWrapper(this)
{
	PG_BackgroundColor = "FillBackColor";
	PG_ForegroundColor = "FillTextColor";
	OwnView();
}

ArpOutlineListView::ArpOutlineListView(BMessage* data)
	: BOutlineListView(data), ArpViewWrapper(data, this)
{
	OwnView();
}

ArpOutlineListView* ArpOutlineListView::Instantiate(BMessage* archive)
{
	if ( validate_instantiation(archive, "ArpOutlineListView") ) 
		return new ArpOutlineListView(archive); 
	return NULL;
}

status_t ArpOutlineListView::Archive(BMessage* data, bool deep)
{
	status_t status = BOutlineListView::Archive(data,false);
	if( status == B_NO_ERROR )
		status = ArpLayoutable::Archive(data,deep);
	if( status == B_NO_ERROR )
		status = data->AddString("class","ArpOutlineListView");
	return status;
}

void ArpOutlineListView::GetPreferredSize(float* w, float* h)
{
	if( w ) *w = 0;
	if( h ) *h = 0;
}

void ArpOutlineListView::MakeFocus(bool focusState)
{
	BOutlineListView::MakeFocus(focusState);
	ArpLayoutable::SetFocusShown(focusState);
}

void ArpOutlineListView::ComputeDimens(void)
{
	ArpViewWrapper::ComputeDimens();
	float min_w=0, min_h=0;
	float pref_w=0, pref_h=0;
	
	int32 num = CountItems();
	int32 max_num = 4;
	for( int32 i=0; i<num; i++ ) {
		BListItem* item = ItemAt(i);
		if( item ) {
			item->Update(this,&PV_BasicFont);
			float w = item->Width();
			float h = item->Height();
			min_h = max(min_h,h);
			pref_w = max(pref_w,w);
			if( max_num > 0 ) pref_h += h;
			max_num--;
		}
	}
	pref_h+=2;
	float fw = PV_BasicFont.StringWidth("WWWW");
	font_height fhs;
	PV_BasicFont.GetHeight(&fhs);
	float fh = fhs.ascent+fhs.descent+fhs.leading;
	min_w = max(min_w,fw);
	min_h = max(min_h,fh);
	pref_w = max(pref_w,min_w);
	pref_h = max(pref_h,min_h);
	cur_dimens.min_width = max(cur_dimens.min_width,min_w);
	cur_dimens.min_height = max(cur_dimens.min_height,min_h);
	cur_dimens.pref_width = max(cur_dimens.pref_width,pref_w);
	cur_dimens.pref_height = max(cur_dimens.pref_height,pref_h);
	cur_dimens.max_width = cur_dimens.max_height = ArpAnySize;
}

/* ------ ArpMenuBar object ------
 */
 
ArpMenuBar::ArpMenuBar(const char* title, menu_layout layout,
						bool resizeToFit)
	:BMenuBar(InitFrame, title, B_FOLLOW_NONE,
				layout, resizeToFit),
	 ArpViewWrapper(this),
	 mLastFrame(Frame())
{
	PG_BackgroundColor = "MenuBackColor";
	PG_ForegroundColor = "MenuTextColor";
	PG_BasicFont = "MenuFont";
	
	OwnView();
	
	// For some reason, the menu doesn't get correctly redrawn
	// when resizing, so do this.  I really don't want to do this,
	// though...
	SetFlags(Flags() | B_FRAME_EVENTS);
}

ArpMenuBar::ArpMenuBar(BMessage* data)
	: BMenuBar(data), ArpViewWrapper(data, this)
{
	OwnView();
}

ArpMenuBar* ArpMenuBar::Instantiate(BMessage* archive)
{
	if ( validate_instantiation(archive, "ArpMenuBar") ) 
		return new ArpMenuBar(archive); 
	return NULL;
}

status_t ArpMenuBar::Archive(BMessage* data, bool deep)
{
	status_t status = BMenuBar::Archive(data,false);
	if( status == B_NO_ERROR )
		status = ArpLayoutable::Archive(data,deep);
	if( status == B_NO_ERROR )
		status = data->AddString("class","ArpMenuBar");
	return status;
}

void ArpMenuBar::FrameResized(float new_width, float new_height) {
	if( Window() ) {
		if( new_width < mLastFrame.Width() ) {
			Invalidate(BRect(new_width-3, Bounds().top,
							 new_width, Bounds().bottom));
		} else if( new_width > mLastFrame.Width() ) {
			Invalidate(BRect(mLastFrame.Width()-3, Bounds().top,
							 mLastFrame.Width(), Bounds().bottom));
		}
	}
	mLastFrame = Frame();
	BMenuBar::FrameResized(new_width, new_height);
}

void ArpMenuBar::copy_font(BMenu* menu, BFont* font)
{
	if( !menu ) return;
	int32 num = menu->CountItems();
	for( int32 i=0; i<num; i++ ) {
		BMenu* child = menu->SubmenuAt(i);
		if( child ) {
			copy_font(child,font);
			child->SetFont(font);
			child->InvalidateLayout();
		}
	}
}

void ArpMenuBar::ComputeDimens(void)
{
	InvalidateLayout();
	// Copy my font to all children
	BFont font;
	GetFont(&font);
	copy_font((BMenu*)this,&font);
	ArpViewWrapper::ComputeDimens();
	cur_dimens.max_width = ArpAnySize;
}

/* ------ ArpMenuField object ------
 */
 
const char* ArpMenuField::MenuFontP = "MenuFont";
const char* ArpMenuField::MenuBackColorP = "MenuBackColor";
const char* ArpMenuField::MenuForeColorP = "MenuForeColor";

ArpMenuField::ArpMenuField(const char* name, const char* label,
						   BMenu* menu, bool fixed_size)
	:BMenuField(InitFrame, name, label, menu, fixed_size, B_FOLLOW_NONE,
				B_WILL_DRAW|B_NAVIGABLE|B_FRAME_EVENTS),
	 ArpViewWrapper(this)
{
	mLabelWidth = 0;
	OwnView();
	SetLayoutName(name);
	PG_MenuBackColor = "MenuBackColor";
	PG_MenuForeColor = "MenuTextColor";
	PG_MenuFont = "MenuFont";
	PV_MenuFont = be_bold_font;
	PV_MenuBackColor = ui_color(B_PANEL_BACKGROUND_COLOR);
	PV_MenuForeColor = ArpColor(0x00,0x00,0x00);
	struct menu_info minfo;
	if( get_menu_info(&minfo) == B_NO_ERROR ) {
		PV_MenuFont.SetFamilyAndStyle(minfo.f_family,minfo.f_style);
		PV_MenuFont.SetSize(minfo.font_size);
		PV_MenuBackColor = minfo.background_color;
	}
}

ArpMenuField::ArpMenuField(BMessage* data)
	: BMenuField(data), ArpViewWrapper(data, this)
{
	mLabelWidth = 0;
	OwnView();
}

ArpMenuField* ArpMenuField::Instantiate(BMessage* archive)
{
	if ( validate_instantiation(archive, "ArpMenuField") ) 
		return new ArpMenuField(archive); 
	return NULL;
}

void ArpMenuField::FrameResized(float width, float height)
{
	//printf("Setting divider label %s width to %f\n", Label(), mLabelWidth);
	SetDivider(mLabelWidth);
}

status_t ArpMenuField::Archive(BMessage* data, bool deep)
{
	status_t status = BMenuField::Archive(data,false);
	if( status == B_NO_ERROR )
		status = ArpLayoutable::Archive(data,deep);
	if( status == B_NO_ERROR )
		status = data->AddString("class","ArpMenuField");
	return status;
}

void ArpMenuField::copy_attrs(BMenu* menu)
{
	if( !menu ) return;
	int32 num = menu->CountItems();
	for( int32 i=0; i<num; i++ ) {
		BMenu* child = menu->SubmenuAt(i);
		if( child ) {
			copy_attrs(child);
			// the menu doesn't seem to see this.
			//child->SetFont(&PV_MenuFont);
			//child->SetHighColor(PV_MenuForeColor);
			//child->SetLowColor(PV_MenuBackColor);
			//child->SetViewColor(PV_MenuBackColor);
			child->InvalidateLayout();
		}
	}
}

void ArpMenuField::DoSetParams(const ArpMessage* g,
								const ArpMessage* p)
{
	ArpViewWrapper::DoSetParams(g,p);
	
	ArpD(cdb << ADH << "ArpMenuField::DoSetParams(), with:" << endl);
	ArpD(if( g ) cdb << ADH << "Globals = " << *g << endl);
	ArpD(if( p ) cdb << ADH << "Params = " << *p << endl);
	
	bool changed = false;
	
	status_t res;
	res = ExtractParam(p,MenuFontP,FFont::FONT_TYPE,
						g,&PG_MenuFont,&PV_MenuFont);
	if( !res ) changed = true;
	res = ExtractParam(p,MenuBackColorP,B_RGB_COLOR_TYPE,
						g,&PG_MenuBackColor,&PV_MenuBackColor);
	if( !res ) changed = true;
	res = ExtractParam(p,MenuForeColorP,B_RGB_COLOR_TYPE,
						g,&PG_MenuForeColor,&PV_MenuForeColor);
	if( !res ) changed = true;
	
	if( changed ) {
		if( MenuBar() ) {
			copy_attrs(MenuBar());
			MenuBar()->SetFont(&PV_BasicFont);
			MenuBar()->SetHighColor(PV_MenuForeColor);
			MenuBar()->SetLowColor(PV_MenuBackColor);
			MenuBar()->SetViewColor(PV_MenuBackColor);
		}
		if( Menu() ) {
			copy_attrs(Menu());
		}
		InvalidateView();
	}
}

bool ArpMenuField::DoParams(ArpMessage& p, const char* name)
{
	bool res = ArpViewWrapper::DoParams(p,name);
	if( !name || !strcmp(name,MenuFontP) ) {
		if( p.AddFont(MenuFontP,&PV_MenuFont) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,MenuBackColorP) ) {
		if( p.AddRGBColor(MenuBackColorP,&PV_MenuBackColor) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,MenuForeColorP) ) {
		if( p.AddRGBColor(MenuForeColorP,&PV_MenuForeColor) ) {
			res = false;
		}
		if( name ) return res;
	}
	return res;
}

void ArpMenuField::ComputeDimens(void)
{
	BMenu* menu = MenuBar();
	BMenu* popup = Menu();
	
	if( menu ) {
		copy_attrs(menu);
		copy_attrs(popup);
		menu->SetFont(&PV_BasicFont);
		menu->InvalidateLayout();
	}
	
	ArpViewWrapper::ComputeDimens();
	
	font_height fhs;
	PV_BasicFont.GetHeight(&fhs);
	const float fh = fhs.ascent+fhs.descent+fhs.leading;
	float fw = PV_BasicFont.StringWidth("WWWW");
	
	float pref_w=0;
	if( popup ) {
		int32 num = popup->CountItems();
		for( int32 i=0; i<num; i++ ) {
			BMenuItem* item = popup->ItemAt(i);
			if( item ) {
				const float w=PV_BasicFont.StringWidth(item->Label());
				ArpD(cdb << ADH << "Dimensions for popup label "
							<< item->Label() << ": " << w << endl);
				if( w > pref_w ) pref_w = w;
			}
		}
	}
	
	cur_dimens.min_height = cur_dimens.pref_height
		= cur_dimens.max_height = fh + 12;
	
	mLabelWidth = (Label() && *Label())
		? PV_BasicFont.StringWidth(Label())	+ PV_BasicFont.StringWidth(" ")
		: 0;
	const float extra_w = mLabelWidth + 20;
	cur_dimens.min_width = (fw < pref_w ? fw : pref_w) + extra_w;
	cur_dimens.max_width = pref_w + extra_w;
	cur_dimens.pref_width = pref_w + extra_w;
	
	//printf("Setting divider label %s width to %f\n", Label(), mLabelWidth);
	SetDivider(mLabelWidth);
}

/* ------ ArpTextControl object ------
 */
 
const char* ArpTextControl::FillBackColorP = "FillBackColor";
const char* ArpTextControl::FillForeColorP = "FillForeColor";
const char* ArpTextControl::FillFontP = "FillFont";
const char* ArpTextControl::MinTextStringP = "MinTextString";
const char* ArpTextControl::PrefTextStringP = "PrefTextString";
const char* ArpTextControl::MaxTextStringP = "MaxTextString";

ArpTextControl::ArpTextControl(const char* name, const char* label,
								const char* text, BMessage* message)
	:BTextControl(InitFrame, name, label, text, message,
					B_FOLLOW_NONE),
	 ArpViewWrapper(this)
{
	OwnView();
	SetLayoutName(name);
	PG_FillBackColor = "FillBackColor";
	PG_FillForeColor = "FillTextColor";
	PG_FillFont = "BasicFont";
	PV_FillBackColor = ArpColor(B_PANEL_BACKGROUND_COLOR)
						.Tint(B_LIGHTEN_MAX_TINT);
	PV_FillForeColor = ArpColor::Black;
	PV_FillFont = PV_BasicFont;
}

ArpTextControl::ArpTextControl(BMessage* data)
	: BTextControl(data), ArpViewWrapper(data, this)
{
	OwnView();
	/* XXX We need to initialize these BEFORE!!!
	PG_FillBackColor = PG_FillForeColor = NULL;
	PV_FillBackColor.red =
		PV_FillBackColor.green =
		PV_FillBackColor.blue = 0xff;
	PV_FillForeColor.red =
		PV_FillForeColor.green =
		PV_FillForeColor.blue = 0x00;
	*/
}

ArpTextControl* ArpTextControl::Instantiate(BMessage* archive)
{
	if ( validate_instantiation(archive, "ArpTextControl") ) 
		return new ArpTextControl(archive); 
	return NULL;
}

status_t ArpTextControl::Archive(BMessage* data, bool deep)
{
	status_t status = BTextControl::Archive(data,false);
	if( status == B_NO_ERROR )
		status = ArpLayoutable::Archive(data,deep);
	if( status == B_NO_ERROR )
		status = data->AddString("class","ArpTextControl");
	return status;
}

void ArpTextControl::AttachedToWindow(void)
{
	BTextControl::AttachedToWindow();
	copy_colors(this);
	copy_font(this);
}

void ArpTextControl::copy_colors(BView* v)
{
	// Make no assumptions about the structure of the children.
	if( !v ) return;
	int32 num = v->CountChildren();
	for( int32 i=0; i<num; i++ ) {
		BView* child = v->ChildAt(i);
		if( child ) {
			copy_colors(child);
			BTextView* text = dynamic_cast<BTextView*>(child);
			if( text ) {
				text->SetFontAndColor(NULL, B_FONT_ALL, &PV_FillForeColor);
			}
			child->SetViewColor(PV_FillBackColor);
			child->SetLowColor(PV_FillBackColor);
			child->SetHighColor(PV_FillForeColor);
			child->Invalidate();
		}
	}
}

void ArpTextControl::DoSetParams(const ArpMessage* g,
								const ArpMessage* p)
{
	ArpViewWrapper::DoSetParams(g,p);
	
	ArpD(cdb << ADH << "ArpTextControl::DoSetParams(), with:" << endl);
	ArpD(if( g ) cdb << ADH << "Globals = " << *g << endl);
	ArpD(if( p ) cdb << ADH << "Params = " << *p << endl);
	
	bool color_changed = false;
	bool layout_changed = false;
	
	status_t res;
	res = ExtractParam(p,FillBackColorP,B_RGB_COLOR_TYPE,
						g,&PG_FillBackColor,&PV_FillBackColor);
	if( !res ) color_changed = true;
	res = ExtractParam(p,FillForeColorP,B_RGB_COLOR_TYPE,
						g,&PG_FillForeColor,&PV_FillForeColor);
	if( !res ) color_changed = true;
	res = ExtractParam(p,FillFontP,FFont::FONT_TYPE,
						g,&PG_FillFont,&PV_FillFont);
	if( !res ) layout_changed = true;
	res = ExtractParam(p,MinTextStringP,B_STRING_TYPE,
						g,NULL,&PV_MinTextString);
	if( !res ) layout_changed = true;
	res = ExtractParam(p,PrefTextStringP,B_STRING_TYPE,
						g,NULL,&PV_PrefTextString);
	if( !res ) layout_changed = true;
	res = ExtractParam(p,MaxTextStringP,B_STRING_TYPE,
						g,NULL,&PV_MaxTextString);
	if( !res ) layout_changed = true;
	
	if( color_changed ) {
		copy_colors((BView*)this);
		InvalidateView();
	}
	if( layout_changed ) {
		copy_font((BView*)this);
		InvalidateDimens();
	}
}

bool ArpTextControl::DoParams(ArpMessage& p, const char* name)
{
	bool res = ArpViewWrapper::DoParams(p,name);
	if( !name || !strcmp(name,FillBackColorP) ) {
		if( p.AddRGBColor(FillBackColorP,&PV_FillBackColor) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,FillForeColorP) ) {
		if( p.AddRGBColor(FillForeColorP,&PV_FillForeColor) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,FillFontP) ) {
		if( p.AddFont(FillFontP,&PV_FillFont) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !PV_MinTextString.IsNull()
			&& (!name || !strcmp(name,MinTextStringP)) ) {
		if( p.AddString(MinTextStringP,PV_MinTextString) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !PV_PrefTextString.IsNull()
			&& (!name || !strcmp(name,PrefTextStringP)) ) {
		if( p.AddString(MinTextStringP,PV_PrefTextString) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !PV_MaxTextString.IsNull()
			&& (!name || !strcmp(name,MaxTextStringP)) ) {
		if( p.AddString(MaxTextStringP,PV_MaxTextString) ) {
			res = false;
		}
		if( name ) return res;
	}
	return res;
}

void ArpTextControl::copy_font(BView* v)
{
	// Make no assumptions about the structure of the children.
	if( !v ) return;
	int32 num = v->CountChildren();
	for( int32 i=0; i<num; i++ ) {
		BView* child = v->ChildAt(i);
		if( child ) {
			copy_font(child);
			BTextView* text = dynamic_cast<BTextView*>(child);
			if( text ) {
				text->SetFontAndColor(&PV_FillFont);
			} else {
				child->SetFont(&PV_FillFont);
			}
		}
	}
}

void ArpTextControl::ComputeDimens(void)
{
	// Copy fill font to all children
	copy_font((BView*)this);
	//ArpViewWrapper::ComputeDimens();
	BRect frm = Frame();
	float vw=0, vh=0;
	ResizeToPreferred();
	GetPreferredSize(&vw,&vh);
	ResizeTo(frm.Width(),frm.Height());
	cur_dimens.min_width = cur_dimens.pref_width = vw;
	cur_dimens.max_width = ArpAnySize;
	cur_dimens.min_height = cur_dimens.pref_height
		= cur_dimens.max_height = vh;
		
	float extraWidth = Divider() + 12;
	//printf("Left: %f, Width: %f, Divider: %f\n",
	//		Frame().left, Frame().Width(), Divider());
	if( !PV_MinTextString.IsNull() ) {
		cur_dimens.min_width = extraWidth
							 + PV_FillFont.StringWidth(PV_MinTextString);
	}
	if( !PV_PrefTextString.IsNull() ) {
		cur_dimens.pref_width = extraWidth
							  + PV_FillFont.StringWidth(PV_PrefTextString);
	}
	if( !PV_MaxTextString.IsNull() ) {
		cur_dimens.max_width = extraWidth
							 + PV_FillFont.StringWidth(PV_MaxTextString);
	}
}
